/** observer.dart --- Class for easily notifying registered functions of events.
 *
 * Copyright (C) 2014  Miguel Guedes
 * 
 * Author: Miguel Guedes <miguel.a.guedes@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see http://www.gnu.org/licenses/.
 *
 * Comments:
 *
 * 
 */


import 'observer_events.dart';
import 'observer_exception.dart';


/** Manager of notifiable events where clients register a function to receive
 * notifications for a particular event.
 *
 * Instantiate this class if you cannot inherit from Observable. */
class Observer
{
  ObserverEvents                on;
  
  Map<String, List<Function> >  _observers;

  
  Observer()
    : _observers = new Map<String, List<Function> >()
  {
    on = new ObserverEvents(this);
  }

  bool add(final String   event,
           final Function fn)
  {
    if(!_observers.containsKey(event))
      _observers[event] = new List<Function>();
    
    if(!_observers[event].contains(fn)) {
      _observers[event].add(fn);
      return true;
    }

    return false;
  }

  bool remove(final String   event,
              final Function fn)
  {
    if(_observers.containsKey(event) && _observers[event].contains(fn)) {
      _observers[event].remove(fn);
      return true;
    }

    return false;
  }

  void noSuchMethod(final Invocation invocation)
  {
    final List args = invocation.positionalArguments;

    /* The only "method" supported is `dispatch'; make sure it is being invoked
     * correctly. */
    if(!invocation.isMethod)
      throw new ObserverException("Invalid invocation");
    else if((new RegExp("\"(.+)\""))
          .firstMatch(invocation.memberName.toString())[1] != 'dispatch')
      throw new ObserverException('No such method: ${invocation.memberName}');
    else if(args.length < 1)
      throw new ObserverException("Event not specified");

    /* Return if no observers registered. */
    if(!_observers.containsKey(args[0]))
      return;
    
    final List<Function> list = _observers[args[0]];
      
    /* An event can be dispatched with no arguments or with exactly only one
     * argument. */
    if(args.length > 1) {
      var arg = args[1];
        
      for(final Function fn in list)
        fn(arg);
    } else {
      for(final Function fn in list)
        fn();
    }        
  }
}